"""Test TimeSeries analysis functions"""

import numpy as np
import pytest

from tape import ColumnMapper, Ensemble, TimeSeries, analysis
from tape.analysis.structure_function.base_argument_container import StructureFunctionArgumentContainer

# pylint: disable=protected-access


@pytest.mark.parametrize("cls", analysis.AnalysisFunction.__subclasses__())
def test_analysis_function(cls, dask_client):
    """
    Test AnalysisFunction child classes
    """
    # We skip child classes with non-trivial constructors
    try:
        obj = cls()
    except TypeError:
        pytest.skip(f"Class {cls} has non-trivial constructor")

    rows = {
        "id": [8001, 8001, 8001, 8001, 8002, 8002, 8002, 8002, 8002],
        "time": [10.1, 10.2, 10.2, 11.1, 11.2, 11.3, 11.4, 15.0, 15.1],
        "band": ["g", "g", "b", "g", "b", "g", "g", "g", "g"],
        "err": [1.0, 2.0, 1.0, 3.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "flux": [1.0, 2.0, 5.0, 3.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    }
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")
    ens = Ensemble(client=False).from_source_dict(rows, column_mapper=cmap)

    assert isinstance(obj.cols(ens), list)
    assert len(obj.cols(ens)) > 0
    assert isinstance(obj.on(ens), list)
    assert len(obj.on(ens)) > 0

    # We assume that there are no mandatory keyword arguments
    result = ens.batch(obj)

    assert len(result) == len(set(rows["id"]))


def test_stetsonj():
    """
    Simple test of StetsonJ function for a known return value
    """

    flux_list = [0, 1, 2, 3, 4]
    test_dict = {
        "time": range(len(flux_list)),
        "flux": flux_list,
        "flux_err": [1] * len(flux_list),
        "band": ["r"] * len(flux_list),
    }
    timseries = TimeSeries()
    test_ts = timseries.from_dict(data_dict=test_dict)
    print("test StetsonJ value is: " + str(test_ts.stetson_J()["r"]))
    assert test_ts.stetson_J()["r"] == 0.8


@pytest.mark.parametrize("lc_id", [None, 1])
def test_sf2_timeseries(lc_id):
    """
    Test of structure function squared function for a known return value
    """

    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]

    test_dict = {
        "time": test_t,
        "flux": test_y,
        "flux_err": test_yerr,
        "band": ["r"] * len(test_y),
    }
    timseries = TimeSeries()
    timseries.meta["id"] = lc_id
    test_series = timseries.from_dict(data_dict=test_dict)

    res = test_series.sf2()

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_timeseries_without_timestamps():
    """
    Test of structure function squared function for a known return value without
    providing timestamp values
    """

    test_t = None
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]

    test_dict = {
        "time": test_t,
        "flux": test_y,
        "flux_err": test_yerr,
        "band": ["r"] * len(test_y),
    }
    timeseries = TimeSeries()
    timeseries.meta["id"] = 1
    test_series = timeseries.from_dict(data_dict=test_dict)
    res = test_series.sf2()

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_timeseries_with_all_none_timestamps():
    """
    Test of structure function squared function for a known return value without
    providing timestamp values
    """

    test_t = [None, None, None, None, None, None, None, None]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]

    test_dict = {
        "time": test_t,
        "flux": test_y,
        "flux_err": test_yerr,
        "band": ["r"] * len(test_y),
    }
    timeseries = TimeSeries()
    timeseries.meta["id"] = 1
    test_series = timeseries.from_dict(data_dict=test_dict)
    res = test_series.sf2()

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_base_case():
    """
    Base test case accessing calc_sf2 directly. Does not make use of TimeSeries
    or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_base_case_time_as_none_array():
    """
    Call calc_sf2 directly. Pass time array of all `None`s.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [None, None, None, None, None, None, None, None]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_base_case_time_as_none_scalar():
    """
    Call calc_sf2 directly. Pass a scalar `None` for time.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = None
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_base_case_string_for_band_to_calc():
    """
    Base test case accessing calc_sf2 directly. Pass a string for band_to_calc.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))
    test_band_to_calc = "r"

    arg_container = StructureFunctionArgumentContainer()
    arg_container.band_to_calc = test_band_to_calc

    res = analysis.calc_sf2(
        time=test_t, flux=test_y, err=test_yerr, band=test_band, lc_id=lc_id, argument_container=arg_container
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_base_case_error_as_scalar():
    """
    Base test case accessing calc_sf2 directly. Provides a scalar value for
    error. Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = 0.1
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.152482, rel=0.001)


def test_sf2_base_case_error_as_none():
    """
    Base test case accessing calc_sf2 directly. Provides `None` for error.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = None
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.172482, rel=0.001)


def test_sf2_no_lightcurve_ids():
    """
    Base test case accessing calc_sf2 directly. Pass no lightcurve ids.
    Does not make use of TimeSeries or Ensemble.
    """
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_no_band_information():
    """
    Base test case accessing calc_sf2 directly. Pass no band information
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)


def test_sf2_least_possible_information():
    """
    Base test case accessing calc_sf2 directly. Pass time as None and flux, but
    nothing else.
    Does not make use of TimeSeries or Ensemble.
    """
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]

    res = analysis.calc_sf2(
        time=None,
        flux=test_y,
    )

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.172482, rel=0.001)


def test_sf2_least_possible_information_constant_flux():
    """
    Base test case accessing calc_sf2 directly. Pass time as None and identical
    flux values, but nothing else.
    Does not make use of TimeSeries or Ensemble.
    """
    test_y = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    res = analysis.calc_sf2(time=None, flux=test_y)

    assert res["dt"][0] == pytest.approx(3.0, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.0, rel=0.001)


# @pytest.mark.skip
def test_sf2_flux_and_band_different_lengths():
    """
    Base test case accessing calc_sf2 directly. Flux and band are different
    lengths. Expect an exception to be raised.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2, 0.3]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_t))

    with pytest.raises(ValueError) as execinfo:
        analysis.calc_sf2(
            time=test_t,
            flux=test_y,
            err=test_yerr,
            band=test_band,
            lc_id=lc_id,
        )

    assert "same length" in str(execinfo.value)


def test_sf2_flux_and_lc_id_different_lengths():
    """
    Base test case accessing calc_sf2 directly. Flux and lc_id are different
    lengths. Expect an exception to be raised.
    Does not make use of TimeSeries or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_t))

    with pytest.raises(ValueError) as execinfo:
        analysis.calc_sf2(
            time=test_t,
            flux=test_y,
            err=test_yerr,
            band=test_band,
            lc_id=lc_id,
        )

    assert "Value of `lc_id` should be `None` or array with the same length" in str(execinfo.value)


def test_create_arg_container_without_arg_container():
    """Base case, no argument container provided, expect a default argument_container
    to be created.
    """

    test_sf_method = "basic"
    default_arg_container = StructureFunctionArgumentContainer()

    output_arg_container = analysis.structurefunction2._create_arg_container_if_needed(test_sf_method, None)
    assert default_arg_container.sf_method == output_arg_container.sf_method
    assert default_arg_container.ignore_timestamps == output_arg_container.ignore_timestamps


def test_create_arg_container_with_arg_container():
    """Base case, with an argument container provided,
    expect that the argument container will be passed through, untouched.
    """

    test_sf_method = "basic"
    default_arg_container = StructureFunctionArgumentContainer()

    # set one property to a non-default value
    default_arg_container.ignore_timestamps = True

    output_arg_container = analysis.structurefunction2._create_arg_container_if_needed(
        test_sf_method, default_arg_container
    )
    assert default_arg_container.sf_method == output_arg_container.sf_method
    assert default_arg_container.ignore_timestamps == output_arg_container.ignore_timestamps


def test_validate_band_with_band_value():
    """Base case where band would be passed in as a non-default value.
    An argument_container is also provided with a different value. We expect
    that the `band` would be the resulting output.
    """

    input_band = ["r"]
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.band = ["b"]

    output_band = analysis.structurefunction2._validate_band(input_band, input_flux, arg_container)

    assert output_band == input_band


def test_validate_band_with_arg_container_band_value():
    """Base case where band would be passed in as a default (`None`) value.
    An argument_container is also provided with an actual value. We expect
    that the `arg_container.band` would be the resulting output.
    """

    input_band = None
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.band = ["b"]

    output_band = analysis.structurefunction2._validate_band(input_band, input_flux, arg_container)

    assert output_band == arg_container.band


def test_validate_band_with_no_input_values():
    """Base case where band is not provided in any location. Expected output is
    an array of 0s equal in length to the input_flux array.
    """

    input_band = None
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    expected_output = np.zeros(len(input_flux), dtype=np.int8)

    output_band = analysis.structurefunction2._validate_band(input_band, input_flux, arg_container)

    assert output_band == expected_output


def test_validate_band_with_band_value_wrong_length():
    """Band will be passed in, but will be a different length than the input flux."""

    input_band = ["r"]
    input_flux = [1, 2]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.band = ["b"]

    with pytest.raises(ValueError) as execinfo:
        analysis.structurefunction2._validate_band(input_band, input_flux, arg_container)

    assert "same length" in str(execinfo.value)


def test_validate_lightcurve_with_lc_value():
    """Base case where lc_id would be passed in as a non-default value.
    An argument_container is also provided with a different value. We expect
    that the `lc_id` would be the resulting output.
    """

    input_lc_id = [100]
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.lc_id = [333]

    output_lc_id = analysis.structurefunction2._validate_lightcurve_id(input_lc_id, input_flux, arg_container)

    assert output_lc_id == input_lc_id


def test_validate_lightcurve_with_arg_container_lc_value():
    """Base case where lc_id would be passed in as a default (`None`) value.
    An argument_container is also provided with an actual value. We expect
    that the `arg_container.lc_id` would be the resulting output.
    """

    input_lc_id = None
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.lc_id = [333]

    output_lc_id = analysis.structurefunction2._validate_lightcurve_id(input_lc_id, input_flux, arg_container)

    assert output_lc_id == arg_container.lc_id


def test_validate_lightcurve_with_no_input_values():
    """Base case where lc_id is not provided in any location. Expected output is
    an array of 0s equal in length to the input_flux array.
    """

    input_lc_id = None
    input_flux = [1]
    arg_container = StructureFunctionArgumentContainer()
    expected_output = np.zeros(len(input_flux), dtype=np.int8)

    output_lc_id = analysis.structurefunction2._validate_lightcurve_id(input_lc_id, input_flux, arg_container)

    assert output_lc_id == expected_output


def test_validate_band_with_band_value_wrong_length():
    """Lightcurve id will be passed in, but will be a different length than the input flux."""

    input_lc_id = [100]
    input_flux = [1, 2]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.lc_id = [333]

    with pytest.raises(ValueError) as execinfo:
        analysis.structurefunction2._validate_band(input_lc_id, input_flux, arg_container)

    assert "same length" in str(execinfo.value)


def test_validate_sf_method_base():
    """Will pass in "basic" and expect "basic" and output."""

    input_sf_method = "basic"
    arg_container = StructureFunctionArgumentContainer()

    output_sf_method = analysis.structurefunction2._validate_sf_method(input_sf_method, arg_container)

    assert input_sf_method == output_sf_method


def test_validate_sf_method_raises_for_unknown_method():
    """Make sure that we raise an exception when an unknown sf_method is provided."""

    input_sf_method = "basic"
    arg_container = StructureFunctionArgumentContainer()
    arg_container.sf_method = "bogus_method"

    with pytest.raises(ValueError) as execinfo:
        analysis.structurefunction2._validate_sf_method(input_sf_method, arg_container)

    assert "Unknown" in str(execinfo.value)


def test_sf2_base_case_macleod_2012():
    """
    Base test case accessing calc_sf2 directly. Uses `MacLeod 2012` SF calculation method.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))
    test_sf_method = "macleod_2012"

    res = analysis.calc_sf2(
        time=test_t, flux=test_y, err=test_yerr, band=test_band, lc_id=lc_id, sf_method=test_sf_method
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.1687, rel=0.001)


def test_sf2_multiple_bands():
    """
    Starts with the base case test, but duplicates the test_t, test_y and test_err
    for a second color band.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2, 1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2, 0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
    ]
    test_band = np.array(
        [
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
        ]
    )

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)
    assert res["dt"][1] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][1] == pytest.approx(0.005365, rel=0.001)


def test_sf2_provide_bins_in_argument_container():
    """
    Base test case accessing calc_sf2 directly. Does not make use of TimeSeries
    or Ensemble.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))

    arg_container = StructureFunctionArgumentContainer()
    arg_container.bins = [0.0, 3.1, 9.0]

    res = analysis.calc_sf2(
        time=test_t, flux=test_y, err=test_yerr, band=test_band, lc_id=lc_id, argument_container=arg_container
    )

    assert res["dt"][0] == pytest.approx(1.7581, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.0103, rel=0.001)
    assert res["dt"][1] == pytest.approx(5.0017, rel=0.001)
    assert res["sf2"][1] == pytest.approx(-0.001216, rel=0.001)


def test_sf2_with_random_sampling_one_lightcurve():
    """
    Base case of passing only one light curve
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.estimate_err = True
    test_arg_container.random_seed = 42

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(3.4475, rel=0.001)
    assert res["sf2"][0] == pytest.approx(-0.0030716, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. Uses a
    predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        4.01,
        5.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.random_seed = 42

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(3.148214, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)
    assert res["dt"][1] == pytest.approx(2.891860, rel=0.001)
    assert res["sf2"][1] == pytest.approx(0.048904, rel=0.001)


def test_sf2_with_unequal_weighting_multiple_lightcurve():
    """
    Passing two light curves with unequal weighting. The first light curve is the
    well tested one, the second has more observations. Uses a
    predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        4.01,
        5.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    # test_band = np.array(["r"] * len(test_y))
    test_band = np.array(
        ["r", "r", "r", "r", "r", "g", "g", "g", "r", "r", "r", "r", "r", "r", "g", "g", "g", "g"]
    )
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = False
    test_arg_container.random_seed = 42
    test_arg_container.bin_count_target = 4

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )
    # change values
    assert res["lc_id"][0] == "1"
    assert res["band"][0] == "g"
    assert res["dt"][0] == pytest.approx(1.2533, rel=0.001)
    assert res["sf2"][0] == pytest.approx(-0.164149, rel=0.001)
    assert res["lc_id"][5] == "2"
    assert res["band"][5] == "r"
    assert res["dt"][5] == pytest.approx(0.8875, rel=0.001)
    assert res["sf2"][5] == pytest.approx(0.0871, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve_multiple_samplings():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. We will
    resample multiple times. Uses a predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        4.01,
        5.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.random_seed = 42
    test_arg_container.calculation_repetitions = 100

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(3.148214, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.005365, rel=0.001)
    assert res["dt"][1] == pytest.approx(2.891860, rel=0.001)
    assert res["sf2"][1] == pytest.approx(0.048904, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve_multiple_samplings_small_bins():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. We will
    resample multiple times. Uses a predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        4.01,
        5.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.random_seed = 42
    test_arg_container.calculation_repetitions = 100
    test_arg_container.bin_count_target = 4

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(0.662500, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.031108, rel=0.001)
    assert res["dt"][7] == pytest.approx(0.643333, rel=0.001)
    assert res["sf2"][7] == pytest.approx(0.070499, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve_multiple_samplings_and_combining():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. We will
    resample multiple times. Also requesting that the results are combined.
    Uses a predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        9.01,
        10.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.estimate_err = True
    test_arg_container.random_seed = 42
    test_arg_container.calculation_repetitions = 3
    test_arg_container.bin_count_target = 4
    test_arg_container.combine = True

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(0.55, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.100116, rel=0.001)
    assert res["1_sigma"][0] == pytest.approx(0.061128, rel=0.001)
    assert res["dt"][9] == pytest.approx(3.035000, rel=0.001)
    assert res["sf2"][9] == pytest.approx(0.039646, rel=0.001)
    assert res["1_sigma"][9] == pytest.approx(0.045012, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve_multiple_samplings_and_combining_non_default_sigma():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. We will
    resample multiple times. Also requesting that the results are combined.
    We will use a non-default value for the upper and lower error quantiles.
    Uses a predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        9.01,
        10.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.estimate_err = True
    test_arg_container.random_seed = 42
    test_arg_container.calculation_repetitions = 3
    test_arg_container.bin_count_target = 4
    test_arg_container.combine = True
    test_arg_container.lower_error_quantile = 0.4
    test_arg_container.upper_error_quantile = 0.6

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(0.55, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.1001165, rel=0.001)
    assert res["1_sigma"][0] == pytest.approx(0.017979, rel=0.001)
    assert res["dt"][9] == pytest.approx(3.035, rel=0.001)
    assert res["sf2"][9] == pytest.approx(0.0396455, rel=0.001)
    assert res["1_sigma"][9] == pytest.approx(0.0132388, rel=0.001)


def test_sf2_with_equal_weighting_multiple_lightcurve_multiple_samplings_small_bins_report_error_seperately():
    """
    Passing two light curves with equal weighting. The first light curve is the
    well tested one, the second is longer so a subsample will be taken. We will
    resample multiple times. Uses a predefined random seed for reproducibility.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
    test_t = [
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        1.11,
        2.23,
        3.45,
        4.01,
        5.67,
        6.32,
        7.88,
        8.2,
        4.01,
        5.67,
    ]
    test_y = [
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.11,
        0.23,
        0.45,
        0.01,
        0.67,
        0.32,
        0.88,
        0.2,
        0.01,
        0.67,
    ]
    test_yerr = [
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.023,
        0.045,
        0.1,
        0.067,
        0.032,
        0.8,
        0.02,
        0.1,
        0.067,
    ]
    test_band = np.array(["r"] * len(test_y))
    test_arg_container = StructureFunctionArgumentContainer()
    test_arg_container.equally_weight_lightcurves = True
    test_arg_container.estimate_err = True
    test_arg_container.random_seed = 42
    test_arg_container.calculation_repetitions = 100
    test_arg_container.bin_count_target = 4
    test_arg_container.report_upper_lower_error_separately = True

    res = analysis.calc_sf2(
        time=test_t,
        flux=test_y,
        err=test_yerr,
        band=test_band,
        lc_id=lc_id,
        argument_container=test_arg_container,
    )

    assert res["dt"][0] == pytest.approx(0.64267, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.040187, rel=0.001)
    assert res["lower_error"][0] == pytest.approx(0.07956, rel=0.001)
    assert res["upper_error"][0] == pytest.approx(0.06126, rel=0.001)
    assert res["dt"][7] == pytest.approx(0.6095, rel=0.001)
    assert res["sf2"][7] == pytest.approx(0.08424, rel=0.001)
    assert res["lower_error"][7] == pytest.approx(0.0962, rel=0.001)
    assert res["upper_error"][7] == pytest.approx(0.051476, rel=0.001)



"""
Auxiliary code for time-series feature extraction with "light-curve"  package
"""

from typing import List

import numpy as np
import pandas as pd
from light_curve.light_curve_ext import _FeatureEvaluator as BaseLightCurveFeature

from tape.analysis.base import AnalysisFunction


__all__ = ["FeatureExtractor", "BaseLightCurveFeature"]


class FeatureExtractor(AnalysisFunction):
    """Apply light-curve package feature extractor to a light curve

    Parameters
    ----------
    feature : light_curve.light_curve_ext._FeatureEvaluator
        Feature extractor to apply, see "light-curve" package for more details.

    Attributes
    ----------
    feature : light_curve.light_curve_ext._FeatureEvaluator
        Feature extractor to apply, see "light-curve" package for more details.
    """

    def __init__(self, feature: BaseLightCurveFeature):
        self.feature = feature

    def cols(self, ens: "Ensemble") -> List[str]:
        return [ens._time_col, ens._flux_col, ens._err_col, ens._band_col]

    def meta(self, ens: "Ensemble") -> pd.DataFrame:
        """Return the schema of the analysis function output.

        It always returns a pandas.DataFrame with the same columns as
        `self.feature.names` and dtype `np.float64`. However, if
        input columns are all single precision floats then the output dtype
        will be `np.float32`.
        """
        return pd.DataFrame(dtype=np.float64, columns=self.feature.names)

    def on(self, ens: "Ensemble") -> List[str]:
        return [ens._id_col]

    def __call__(self, time, flux, err, band, *, band_to_calc: str, **kwargs) -> pd.DataFrame:
        """
        Apply a feature extractor to a light curve, concatenating the results over
        all bands.

        Parameters
        ----------
        time : `numpy.ndarray`
            Time values
        flux : `numpy.ndarray`
            Brightness values, flux or magnitudes
        err : `numpy.ndarray`
            Errors for "flux"
        band : `numpy.ndarray`
            Passband names.
        band_to_calc : `str` or `int` or `None`
            Name of the passband to calculate features for, usually a string
            like "g" or "r", or an integer. If None, then features are
            calculated for all sources - band is ignored.
        **kwargs : `dict`
            Additional keyword arguments to pass to the feature extractor.

        Returns
        -------
        features : pandas.DataFrame
            Feature values for each band, dtype is a common type for input arrays.
        """

        # Select passband to calculate
        if band_to_calc is not None:
            band_mask = band == band_to_calc
            time, flux, err = (a[band_mask] for a in (time, flux, err))

        # Sort inputs by time if not already sorted
        if not kwargs.get("sorted", False):
            sort_idx = np.argsort(time)
            time, flux, err, band = (a[sort_idx] for a in (time, flux, err, band))
            # Now we can update the kwargs for better performance
            kwargs = kwargs.copy()
            kwargs["sorted"] = True

        # Convert the numerical arrays to a common dtype
        dtype = np.find_common_type([a.dtype for a in (time, flux, err)], [])
        time, flux, err = (a.astype(dtype) for a in (time, flux, err))

        values = self.feature(time, flux, err, **kwargs)

        series = pd.Series(dict(zip(self.feature.names, values)))
        return series



from .base import AnalysisFunction  # noqa
from .feature_extractor import FeatureExtractor  # noqa
from .light_curve import LightCurve  # noqa
from .stetsonj import *  # noqa
from .structurefunction2 import *  # noqa



import numpy as np


class LightCurve:
    """This base class is meant to support various analysis routines and be
    extended as needed. (Hence it's location in the `analysis` package.)

    The base class ensures that the data for a single lightcurve is well formed.
    Namely that the input data is all of the same length, with NaN's removed and
    that there are enough observations to perform a given analysis.
    """

    def __init__(
        self, times: np.ndarray, fluxes: np.ndarray, errors: np.ndarray, minimum_observations: int = 0
    ):
        """Calls the processing functions to clean and validate the input data.
        All of the input arrays should have the same length.

        Parameters
        ----------
        times : np.ndarray
            The timestamps values of the data.
        fluxes : np.ndarray
            The flux/magnitude values of the data.
        errors : np.ndarray
            The error in the measurements.
        minimum_observations : int, optional
            The minimum number of data points needed to successfully process
            the lightcurve, by default 0
        """
        # The primary data
        self._times = times
        self._fluxes = fluxes
        self._errors = errors
        self._minimum_observations = minimum_observations

        self._process_input_data()

    def _process_input_data(self):
        """Cleaning and validation occurs here, ideally by calling
        sub-methods for specific checks and cleaning tasks.
        """
        self._check_input_data_size_is_equal()  # make sure arrays the same length
        self._filter_nans()
        self._check_input_data_length_is_sufficient()  # make sure we have enough values still

    def _filter_nans(self):
        """Mask out any NaN values from time, flux and error arrays"""

        t_mask = np.isnan(self._times)
        f_mask = np.isnan(self._fluxes)
        e_mask = np.isnan(self._errors)

        # np.logical_or takes at most 2 arguments, using reduce will collapse
        # the first two arguments into one, and apply it to the third.
        nan_mask = np.logical_or.reduce((t_mask, f_mask, e_mask))

        self._times = self._times[~nan_mask]
        self._fluxes = self._fluxes[~nan_mask]
        self._errors = self._errors[~nan_mask]

    def _check_input_data_size_is_equal(self):
        """Make sure that the three input np.arrays have the same size"""
        if self._times.size != self._fluxes.size or self._times.size != self._errors.size:
            raise ValueError("Input np.arrays are expected to have the same size.")

    def _check_input_data_length_is_sufficient(self):
        """Make sure that we have enough data after cleaning and filtering
        to be able to perform Structure Function calculations.
        """
        if self._times.size < self._minimum_observations:
            raise ValueError("Too few observations provided to create `LightCurve`.")



"""
Contains the base class for analysis functions.
"""

from abc import ABC, abstractmethod
from typing import Callable, List

import pandas as pd


class AnalysisFunction(ABC, Callable):
    """Base class for analysis functions.

    Analysis functions are functions that take few arrays representing
    an object and return a single pandas.Series representing the result.

    Methods
    -------
    cols(ens) -> List[str]
        Return the columns that the analysis function takes as input.
    meta(ens) -> pd.DataFrame
        Return the metadata pandas.DataFrame required by Dask to pre-build
        a computation graph. It is basically the schema for calculate() method
        output.
    on(ens) -> List[str]
        Return the columns to group source table by.
        Typically, `[ens._id_col]`.
    __call__(*cols, **kwargs)
        Calculate the analysis function.
    """

    @abstractmethod
    def cols(self, ens: "Ensemble") -> List[str]:
        """
        Return the column names that the analysis function takes as input.

        Parameters
        ----------
        ens : Ensemble
            The ensemble object, it could be required to get column names of
            the "special" columns like `ens._time_col` or `ens._err_col`.

        Returns
        -------
        List[str]
            The column names to select and pass to .calculate() method.
            For example `[ens._time_col, ens._flux_col]`.
        """
        raise NotImplementedError

    @abstractmethod
    def meta(self, ens: "Ensemble"):
        """
        Return the schema of the analysis function output.

        Parameters
        ----------
        ens : Ensemble
            The ensemble object.

        Returns
        -------
        pd.DataFrame or (str, dtype) tuple or {str: dtype} dictionary
            Dask meta, for example
            `pd.DataFrame(columns=['x', 'y'], dtype=float)`.
        """
        raise NotImplementedError

    @abstractmethod
    def on(self, ens: "Ensemble") -> List[str]:
        """Return the columns to group source table by.

        Parameters
        ----------
        ens : Ensemble
            The ensemble object.

        Returns
        --------
        List[str]
            The column names to group by. Typically, `[ens._id_col]`.
        """
        return [ens._id_col]

    @abstractmethod
    def __call__(self, *cols, **kwargs):
        """Calculate the analysis function.

        Parameters
        ----------
        *cols : array_like
            The columns to calculate the analysis function on. It must be
            consistent with .cols(ens) output.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        pd.Series or pd.DataFrame or array or value
            The result, it must be consistent with .meta() output.
        """
        raise NotImplementedError



"""Test fixtures for Ensemble manipulations"""

import numpy as np
import pandas as pd
import dask.dataframe as dd
import pytest
import tape
import lsdb

from dask.distributed import Client

from tape import Ensemble
from tape.utils import ColumnMapper


@pytest.fixture
def create_test_rows():
    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])

    rows = {
        "id": 8000 + (np.arange(num_points) % 5),
        "time": np.arange(num_points),
        "flux": np.arange(num_points) % len(all_bands),
        "band": np.repeat(all_bands, num_points / len(all_bands)),
        "err": 0.1 * (np.arange(num_points) % 10),
        "count": np.arange(num_points),
        "something_else": np.full(num_points, None),
    }

    return rows


@pytest.fixture
def create_test_column_mapper():
    return ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")


@pytest.fixture
@pytest.mark.parametrize("create_test_rows", [("create_test_rows")])
def create_test_source_table(create_test_rows, npartitions=1):
    return dd.from_dict(create_test_rows, npartitions)


@pytest.fixture
def create_test_object_table(npartitions=1):
    n_obj = 5
    id = 8000 + np.arange(n_obj)
    name = id.astype(str)
    return dd.from_dict(dict(id=id, name=name), npartitions)


# pylint: disable=redefined-outer-name
@pytest.fixture
@pytest.mark.parametrize(
    "create_test_source_table, create_test_column_mapper",
    [("create_test_source_table", "create_test_column_mapper")],
)
def read_dask_dataframe_ensemble(create_test_source_table, create_test_column_mapper):
    return tape.read_dask_dataframe(
        dask_client=False,
        source_frame=create_test_source_table,
        column_mapper=create_test_column_mapper,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
@pytest.mark.parametrize(
    "create_test_source_table, create_test_object_table, create_test_column_mapper",
    [("create_test_source_table", "create_test_object_table", "create_test_column_mapper")],
)
def read_dask_dataframe_with_object_ensemble(
    create_test_source_table, create_test_object_table, create_test_column_mapper
):
    return tape.read_dask_dataframe(
        source_frame=create_test_source_table,
        object_frame=create_test_object_table,
        dask_client=False,
        column_mapper=create_test_column_mapper,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
@pytest.mark.parametrize(
    "create_test_rows, create_test_column_mapper", [("create_test_rows", "create_test_column_mapper")]
)
def read_pandas_ensemble(create_test_rows, create_test_column_mapper):
    return tape.read_pandas_dataframe(
        source_frame=pd.DataFrame(create_test_rows),
        column_mapper=create_test_column_mapper,
        dask_client=False,
        npartitions=1,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
@pytest.mark.parametrize(
    "create_test_rows, create_test_column_mapper", [("create_test_rows", "create_test_column_mapper")]
)
def read_pandas_with_object_ensemble(create_test_rows, create_test_column_mapper):
    n_obj = 5
    id = 8000 + np.arange(n_obj)
    name = id.astype(str)
    object_table = pd.DataFrame(dict(id=id, name=name))

    """Create an Ensemble from pandas dataframes."""
    return tape.read_pandas_dataframe(
        dask_client=False,
        source_frame=pd.DataFrame(create_test_rows),
        object_frame=object_table,
        column_mapper=create_test_column_mapper,
        npartitions=1,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble_with_client(dask_client):
    """Create an Ensemble from parquet data with a dask client."""
    return tape.read_parquet(
        source_file="tests/tape_tests/data/source/test_source.parquet",
        object_file="tests/tape_tests/data/object/test_object.parquet",
        dask_client=dask_client,
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble():
    """Create an Ensemble from parquet data."""
    return tape.read_parquet(
        source_file="tests/tape_tests/data/source/test_source.parquet",
        object_file="tests/tape_tests/data/object/test_object.parquet",
        dask_client=False,
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble_from_source():
    """Create an Ensemble from parquet data, with object file withheld."""
    return tape.read_parquet(
        source_file="tests/tape_tests/data/source/test_source.parquet",
        dask_client=False,
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble_with_column_mapper():
    """Create an Ensemble from parquet data, with object file withheld."""
    colmap = ColumnMapper().assign(
        id_col="ps1_objid",
        time_col="midPointTai",
        flux_col="psFlux",
        err_col="psFluxErr",
        band_col="filterName",
    )

    return tape.read_parquet(
        source_file="tests/tape_tests/data/source/test_source.parquet",
        column_mapper=colmap,
        dask_client=False,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble_with_known_column_mapper():
    """Create an Ensemble from parquet data, with object file withheld."""
    colmap = ColumnMapper().use_known_map("ZTF")

    return tape.read_parquet(
        source_file="tests/tape_tests/data/source/test_source.parquet",
        column_mapper=colmap,
        dask_client=False,
    )


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_parquet_ensemble_from_hipscat():
    """Create an Ensemble from a hipscat/hive-style directory."""

    colmap = ColumnMapper(
        id_col="_hipscat_index",
        time_col="mjd",
        flux_col="mag",
        err_col="Norder",  # no error column...
        band_col="band",
    )

    return tape.read_hipscat(
        "tests/tape_tests/data/small_sky_hipscat/small_sky_source_catalog",
        "tests/tape_tests/data/small_sky_hipscat/small_sky_object_catalog",
        column_mapper=colmap,
        object_index="id",
        source_index="object_id",
        dask_client=False,
    )


@pytest.fixture(scope="package", name="dask_client")
def dask_client():
    """Create a single client for use by all unit test cases."""
    client = Client()
    yield client
    client.close()


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_with_client(dask_client):
    """Create an Ensemble from parquet data without a dask client."""
    ens = Ensemble(client=dask_client)
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        "tests/tape_tests/data/object/test_object.parquet",
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )

    return ens


@pytest.fixture
def parquet_files_and_ensemble_with_client(dask_client):
    """Create an Ensemble from parquet data without a dask client."""
    ens = Ensemble(client=dask_client)
    source_file = "tests/tape_tests/data/source/test_source.parquet"
    object_file = "tests/tape_tests/data/object/test_object.parquet"
    colmap = ColumnMapper().assign(
        id_col="ps1_objid",
        time_col="midPointTai",
        flux_col="psFlux",
        err_col="psFluxErr",
        band_col="filterName",
    )
    ens = ens.from_parquet(source_file, object_file, column_mapper=colmap)
    return ens, source_file, object_file, colmap


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        "tests/tape_tests/data/object/test_object.parquet",
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_partition_size():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        "tests/tape_tests/data/object/test_object.parquet",
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
        partition_size="1MB",
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_with_divisions():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        "tests/tape_tests/data/object/test_object.parquet",
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
        sort=True,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_from_source():
    """Create an Ensemble from parquet data, with object file withheld."""
    ens = Ensemble(client=False)
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        id_col="ps1_objid",
        time_col="midPointTai",
        band_col="filterName",
        flux_col="psFlux",
        err_col="psFluxErr",
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_with_column_mapper():
    """Create an Ensemble from parquet data, with object file withheld."""
    ens = Ensemble(client=False)

    colmap = ColumnMapper().assign(
        id_col="ps1_objid",
        time_col="midPointTai",
        flux_col="psFlux",
        err_col="psFluxErr",
        band_col="filterName",
    )
    ens.from_parquet(
        "tests/tape_tests/data/source/test_source.parquet",
        column_mapper=colmap,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def parquet_ensemble_from_hipscat():
    """Create an Ensemble from a hipscat/hive-style directory."""
    ens = Ensemble(client=False)

    colmap = ColumnMapper(
        id_col="_hipscat_index",
        time_col="mjd",
        flux_col="mag",
        err_col="Norder",  # no error column...
        band_col="band",
    )

    ens.from_hipscat(
        "tests/tape_tests/data/small_sky_hipscat/small_sky_source_catalog",
        "tests/tape_tests/data/small_sky_hipscat/small_sky_object_catalog",
        column_mapper=colmap,
        object_index="id",
        source_index="object_id",
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def ensemble_from_lsdb():
    """Create a dask dataframe from LSDB catalogs"""
    object_cat = lsdb.read_hipscat("tests/tape_tests/data/small_sky_hipscat/small_sky_object_catalog")
    source_cat = lsdb.read_hipscat("tests/tape_tests/data/small_sky_hipscat/small_sky_source_catalog")

    # Pain points: Suffixes here are a bit annoying, and I'd ideally want just the source columns (especially at scale)
    # We do this to get the source catalog indexed by the objects hipscat index
    joined_source_cat = object_cat.join(
        source_cat, left_on="id", right_on="object_id", suffixes=("_object", "")
    )

    colmap = ColumnMapper(
        id_col="_hipscat_index",
        time_col="mjd",
        flux_col="mag",
        err_col="Norder",  # no error column...
        band_col="band",
    )

    ens = Ensemble(client=False)

    # We just avoid needing to invoke the ._ddf property from the catalogs
    ens.from_lsdb(joined_source_cat, object_cat, column_mapper=colmap)

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def read_ensemble_from_lsdb():
    """Create a dask dataframe from LSDB catalogs"""
    object_cat = lsdb.read_hipscat("tests/tape_tests/data/small_sky_hipscat/small_sky_object_catalog")
    source_cat = lsdb.read_hipscat("tests/tape_tests/data/small_sky_hipscat/small_sky_source_catalog")

    # Pain points: Suffixes here are a bit annoying, and I'd ideally want just the source columns (especially at scale)
    # We do this to get the source catalog indexed by the objects hipscat index
    joined_source_cat = object_cat.join(
        source_cat, left_on="id", right_on="object_id", suffixes=("_object", "")
    )

    colmap = ColumnMapper(
        id_col="_hipscat_index",
        time_col="mjd",
        flux_col="mag",
        err_col="Norder",  # no error column...
        band_col="band",
    )

    # We just avoid needing to invoke the ._ddf property from the catalogs
    ens = tape.read_lsdb(joined_source_cat, object_cat, column_mapper=colmap, dask_client=False)

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def dask_dataframe_ensemble():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)

    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])
    rows = {
        "id": 8000 + (np.arange(num_points) % 5),
        "time": np.arange(num_points),
        "flux": np.arange(num_points) % len(all_bands),
        "band": np.repeat(all_bands, num_points / len(all_bands)),
        "err": 0.1 * (np.arange(num_points) % 10),
        "count": np.arange(num_points),
        "something_else": np.full(num_points, None),
    }
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")

    ens.from_dask_dataframe(
        source_frame=dd.from_dict(rows, npartitions=1),
        column_mapper=cmap,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def dask_dataframe_ensemble_partition_size():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)

    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])
    rows = {
        "id": 8000 + (np.arange(num_points) % 5),
        "time": np.arange(num_points),
        "flux": np.arange(num_points) % len(all_bands),
        "band": np.repeat(all_bands, num_points / len(all_bands)),
        "err": 0.1 * (np.arange(num_points) % 10),
        "count": np.arange(num_points),
        "something_else": np.full(num_points, None),
    }
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")

    ens.from_dask_dataframe(
        source_frame=dd.from_dict(rows, npartitions=1),
        column_mapper=cmap,
        partition_size="1MB",
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def dask_dataframe_with_object_ensemble():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)

    n_obj = 5
    id = 8000 + np.arange(n_obj)
    name = id.astype(str)
    object_table = dd.from_dict(
        dict(id=id, name=name),
        npartitions=1,
    )

    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])
    source_table = dd.from_dict(
        {
            "id": 8000 + (np.arange(num_points) % n_obj),
            "time": np.arange(num_points),
            "flux": np.arange(num_points) % len(all_bands),
            "band": np.repeat(all_bands, num_points / len(all_bands)),
            "err": 0.1 * (np.arange(num_points) % 10),
            "count": np.arange(num_points),
            "something_else": np.full(num_points, None),
        },
        npartitions=1,
    )
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")

    ens.from_dask_dataframe(
        source_frame=source_table,
        object_frame=object_table,
        column_mapper=cmap,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def pandas_ensemble():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)

    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])
    rows = {
        "id": 8000 + (np.arange(num_points) % 5),
        "time": np.arange(num_points),
        "flux": np.arange(num_points) % len(all_bands),
        "band": np.repeat(all_bands, num_points / len(all_bands)),
        "err": 0.1 * (np.arange(num_points) % 10),
        "count": np.arange(num_points),
        "something_else": np.full(num_points, None),
    }
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")

    ens.from_pandas(
        pd.DataFrame(rows),
        column_mapper=cmap,
        npartitions=1,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def pandas_with_object_ensemble():
    """Create an Ensemble from parquet data."""
    ens = Ensemble(client=False)

    n_obj = 5
    id = 8000 + np.arange(n_obj)
    name = id.astype(str)
    object_table = pd.DataFrame(
        dict(id=id, name=name),
    )

    num_points = 1000
    all_bands = np.array(["r", "g", "b", "i"])
    source_table = pd.DataFrame(
        {
            "id": 8000 + (np.arange(num_points) % n_obj),
            "time": np.arange(num_points),
            "flux": np.arange(num_points) % len(all_bands),
            "band": np.repeat(all_bands, num_points / len(all_bands)),
            "err": 0.1 * (np.arange(num_points) % 10),
            "count": np.arange(num_points),
            "something_else": np.full(num_points, None),
        },
    )
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")

    ens.from_pandas(
        source_frame=source_table,
        object_frame=object_table,
        column_mapper=cmap,
        npartitions=1,
    )

    return ens


# pylint: disable=redefined-outer-name
@pytest.fixture
def ensemble_from_source_dict():
    """Create an Ensemble from a source dict, returning the ensemble and the source dict."""
    ens = Ensemble(client=False)

    # Create some fake data with two IDs (8001, 8002), two bands ["g", "b"]
    # a few time steps, flux, and data for zero point calculations.
    source_dict = {
        "id": [8001, 8001, 8002, 8002, 8002],
        "time": [1, 2, 3, 4, 5],
        "flux": [30.5, 70, 80.6, 30.2, 60.3],
        "zp_mag": [25.0, 25.0, 25.0, 25.0, 25.0],
        "zp_flux": [10**10, 10**10, 10**10, 10**10, 10**10],
        "error": [10, 10, 10, 10, 10],
        "band": ["g", "g", "b", "b", "b"],
    }
    # map flux_col to one of the flux columns at the start
    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="error", band_col="band")
    ens.from_source_dict(source_dict, column_mapper=cmap)

    return ens, source_dict



import numpy as np
import pytest

from tape.analysis.light_curve import LightCurve


def test_lightcurve_creation():
    """Test the creation of a lightcurve, check that the number of
    expected delta values equals expectations.
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    lc = LightCurve(test_t, test_y, test_yerr)

    assert test_t.size == lc._times.size


def test_lightcurve_removed_nans():
    """Check that NaN values are removed and the data has the expected length."""
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, np.nan])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, np.nan, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, np.nan, 0.8, 0.02])

    lc = LightCurve(test_t, test_y, test_yerr)

    assert lc._times.size == test_t.size - 3


def test_lightcurve_validate_same_length():
    """Make sure that we raise a ValueError if the input np.arrays don't
    have the same length
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    with pytest.raises(ValueError) as excinfo:
        _ = LightCurve(test_t, test_y, test_yerr)

    assert "Input np.arrays are expected to have the same size." in str(excinfo.value)


def test_lightcurve_minimum_length_raises():
    """Make sure that we raise a ValueError if the input np.arrays don't
    have enough data.
    """
    test_t = np.array([])
    test_y = np.array([])
    test_yerr = np.array([])

    with pytest.raises(ValueError) as excinfo:
        _ = LightCurve(test_t, test_y, test_yerr, minimum_observations=1)

    assert "Too few observations provided to create `LightCurve`." in str(excinfo.value)


def test_lightcurve_minimum_length_raises_after_nan_removal():
    """Make sure that we raise a ValueError if the input np.arrays don't
    have enough data after we remove NaN values.
    """
    test_t = np.array([np.nan])
    test_y = np.array([np.nan])
    test_yerr = np.array([np.nan])

    with pytest.raises(ValueError) as excinfo:
        _ = LightCurve(test_t, test_y, test_yerr, minimum_observations=1)

    assert "Too few observations provided to create `LightCurve`." in str(excinfo.value)



from collections import namedtuple


class ColumnMapper:
    """Maps columns from a given dataset into known ensemble column"""

    def __init__(
        self,
        id_col=None,
        time_col=None,
        flux_col=None,
        err_col=None,
        band_col=None,
    ):
        """

        Parameters
        ----------
        id_col: 'str', optional
            Identifies which column contains the Object IDs
        time_col: 'str', optional
            Identifies which column contains the time information
        flux_col: 'str', optional
            Identifies which column contains the flux/magnitude information
        err_col: 'str', optional
            Identifies which column contains the flux/mag error information
        band_col: 'str', optional
            Identifies which column contains the band information

        Returns
        -------
        ColumnMapper object
        """

        Column = namedtuple("Column", ["name", "is_required"])

        self.map = {
            "id_col": id_col,
            "time_col": time_col,
            "flux_col": flux_col,
            "err_col": err_col,
            "band_col": band_col,
        }

        self.required = [
            Column("id_col", True),
            Column("time_col", True),
            Column("flux_col", True),
            Column("err_col", True),
            Column("band_col", True),
        ]

        self.known_maps = {"ZTF": ZTFColumnMapper, "PS1": PS1ColumnMapper}

    def _set_known_map(self, hipscat=True):
        """Must be defined in a known map class"""
        raise NotImplementedError

    def use_known_map(self, map_id, hipscat=True):
        """Use a known mapping scheme

        Parameters
        ----------
        map_id: 'str'
            Identifies which mapping scheme to use
        hipscat: 'bool'
            Indicates whether the data is in hipscat format or not, which will
            affect the chosen ID column (_hipscat_index will be used when
            hipscat is true. True by default.

        Returns
        -------
        A ColumnMapper subclass object dependent on the map_id provided,
        ZTFColumnMapper in the case of "ZTF" for example

        """
        if map_id.upper() in self.known_maps:
            return self.known_maps[map_id.upper()]()._set_known_map(hipscat=hipscat)
        else:
            raise ValueError(f'Unknown Mapping: "{map_id}"')

    def is_ready(self, show_needed=False):
        """shows whether the ColumnMapper has all critical columns assigned

        Parameters
        ----------
        show_needed: 'bool', optional
            Indicates whether to also return a list of missing columns

        Returns
        -------
        `bool` or tuple of (bool, list) dependent on show_needed parameter

        """

        # Grab required column keys
        required_keys = [col.name for col in self.required if col.is_required]

        # Check the map for assigned keys
        ready = True
        needed = []
        for key in required_keys:
            if self.map[key] is None:
                needed.append(key)
                ready = False

        if show_needed:
            return (ready, needed)
        else:
            return ready

    def assign(
        self,
        id_col=None,
        time_col=None,
        flux_col=None,
        err_col=None,
        band_col=None,
    ):
        """Updates a given set of columns

        Parameters
        ----------
        id_col: 'str', optional
            Identifies which column contains the Object IDs
        time_col: 'str', optional
            Identifies which column contains the time information
        flux_col: 'str', optional
            Identifies which column contains the flux/magnitude information
        err_col: 'str', optional
            Identifies which column contains the flux/mag error information
        band_col: 'str', optional
            Identifies which column contains the band information
        nobs_col: list of 'str', optional
            Identifies which columns contain number of observations for each
            band, if available in the input object file
        nobs_tot_col: 'str', optional
            Identifies which column contains the total number of observations,
            if available in the input object file
        """
        assign_map = {
            "id_col": id_col,
            "time_col": time_col,
            "flux_col": flux_col,
            "err_col": err_col,
            "band_col": band_col,
        }

        for item in assign_map.items():
            if item[1] is not None:
                self.map[item[0]] = item[1]

        return self


class ZTFColumnMapper(ColumnMapper):
    """This class establishs a known mapping to Zwicky Transient Facility (ZTF)
    catalog data columns"""

    def _set_known_map(self, hipscat=True):
        """sets the map to a pre-defined ZTF mapping"""
        if hipscat:
            id_col = "_hipscat_index"
        else:
            id_col = "ps1_objid"
        self.map = {
            "id_col": id_col,
            "time_col": "mjd",
            "flux_col": "mag",
            "err_col": "magerr",
            "band_col": "band",
        }
        return self


class PS1ColumnMapper(ColumnMapper):
    """This class establishs a known mapping to Pan-STARRs (PS1)
    catalog data columns"""

    def _set_known_map(self, hipscat=True):
        """sets the map to a pre-defined PS1 mapping"""
        if hipscat:
            id_col = "_hipscat_index"
        else:
            id_col = "objID"
        self.map = {
            "id_col": id_col,
            "time_col": "obsTime",
            "flux_col": "apFlux",
            "err_col": "apFluxErr",
            "band_col": "filterID",
        }
        return self



from .analysis import *  # noqa
from .ensemble import *  # noqa
from .ensemble_frame import *  # noqa
from .timeseries import *  # noqa
from .ensemble_readers import *  # noqa
from ._version import __version__  # noqa



from typing import Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from tape.analysis.base import AnalysisFunction


__all__ = ["calc_stetson_J", "StetsonJ"]


class StetsonJ(AnalysisFunction):
    """Compute the StetsonJ statistic on data from one or several bands"""

    def cols(self, ens: "Ensemble") -> List[str]:
        return [ens._flux_col, ens._err_col, ens._band_col]

    def meta(self, ens: "Ensemble"):
        return "stetsonJ", float

    def on(self, ens: "Ensemble") -> List[str]:
        return [ens._id_col]

    def __call__(
        self,
        flux: np.ndarray,
        err: np.ndarray,
        band: np.ndarray,
        *,
        band_to_calc: Union[str, Iterable[str], None] = None,
        check_nans: bool = False,
    ):
        """Compute the StetsonJ statistic on data from one or several bands

        Parameters
        ----------
        flux : `numpy.ndarray` (N,)
            Array of flux/magnitude measurements
        err : `numpy.ndarray` (N,)
            Array of associated flux/magnitude errors
        band : `numpy.ndarray` (N,)
            Array of associated band labels
        band_to_calc : `str` or `list` of `str`
            Bands to calculate StetsonJ on. Single band descriptor, or list
            of such descriptors.
        check_nans : `bool`
            Boolean to run a check for NaN values and filter them out.

        Returns
        -------
        stetsonJ : `dict`
            StetsonJ statistic for each of input bands.

        Note
        ----------
        In case that no value for `band_to_calc` is passed, the function is
        executed on all available bands in `band`.
        """

        # NaN filtering
        if check_nans:
            f_mask = np.isnan(flux)
            e_mask = np.isnan(err)  # always mask out nan errors?
            nan_mask = np.logical_or(f_mask, e_mask)

            flux = flux[~nan_mask]
            err = err[~nan_mask]
            band = band[~nan_mask]

        unq_band = np.unique(band)

        if band_to_calc is None:
            band_to_calc = unq_band
        if isinstance(band_to_calc, str):
            band_to_calc = [band_to_calc]

        assert isinstance(band_to_calc, Iterable) is True

        stetsonJ = {}
        for b in band_to_calc:
            if b in unq_band:
                mask = band == b
                fluxes = flux[mask]
                errors = err[mask]
                stetsonJ[b] = _stetson_J_single(fluxes, errors)
            else:
                stetsonJ[b] = np.nan

        return stetsonJ


calc_stetson_J = StetsonJ()
calc_stetson_J.__doc__ = StetsonJ.__call__.__doc__


def _stetson_J_single(fluxes, errors):
    """Compute the single band stetsonJ statistic.

    Parameters
    ----------
    fluxes : `numpy.ndarray` (N,)
        Lightcurve flux values.
    errors : `numpy.ndarray` (N,)
        Errors on the lightcurve fluxes.

    Returns
    -------
    stetsonJ : `float`
        StetsonJ statistic

    References
    ----------
    .. [1] Stetson, P. B., "On the Automatic Determination of Light-Curve
    Parameters for Cepheid Variables", PASP, 108, 851S, 1996

    Note
    ----------
    Taken from
    https://github.com/lsst/meas_base/blob/main/python/lsst/meas/base/diaCalculationPlugins.py
    Using the function on random gaussian distribution gives result of -0.2
    instead of expected result of 0?
    """

    n_points = len(fluxes)
    if n_points <= 1:
        return np.nan
    flux_mean = _stetson_J_mean(fluxes, errors)
    delta_val = np.sqrt(n_points / (n_points - 1)) * (fluxes - flux_mean) / errors
    p_k = delta_val**2 - 1
    return np.mean(np.sign(p_k) * np.sqrt(np.fabs(p_k)))


def _stetson_J_mean(values, errors, mean=None, alpha=2.0, beta=2.0, n_iter=20, tol=1e-6):
    """Compute the stetson mean of the fluxes which down-weights outliers.

    Weighted biased on an error weighted difference scaled by a constant
    (1/``a``) and raised to the power beta. Higher betas more harshly
    penalize outliers and ``a`` sets the number of sigma where a weighted
    difference of 1 occurs.

    Parameters
    ----------
    values : `numpy.dnarray`, (N,)
        Input values to compute the mean of.
    errors : `numpy.ndarray`, (N,)
        Errors on the input values.
    mean : `float`
        Starting mean value or None.
    alpha : `float`
        Scalar down-weighting of the fractional difference. lower->more
        clipping. (Default value is 2.)
    beta : `float`
        Power law slope of the used to down-weight outliers. higher->more
        clipping. (Default value is 2.)
    n_iter : `int`
        Number of iterations of clipping.
    tol : `float`
        Fractional and absolute tolerance goal on the change in the mean
        before exiting early. (Default value is 1e-6)

    Returns
    -------
    mean : `float`
        Weighted stetson mean result.

    References
    ----------
    .. [1] Stetson, P. B., "On the Automatic Determination of Light-Curve
    Parameters for Cepheid Variables", PASP, 108, 851S, 1996

    Note
    ----------
    Taken from
    https://github.com/lsst/meas_base/blob/main/python/lsst/meas/base/diaCalculationPlugins.py
    """
    n_points = len(values)
    n_factor = np.sqrt(n_points / (n_points - 1))
    inv_var = 1 / errors**2
    if mean is None:
        mean = np.average(values, weights=inv_var)
    for iter_idx in range(n_iter):
        chi = np.fabs(n_factor * (values - mean) / errors)
        tmp_mean = np.average(values, weights=inv_var / (1 + (chi / alpha) ** beta))
        diff = np.fabs(tmp_mean - mean)
        mean = tmp_mean
        if mean == 0:  # catch divide by zero in diff / mean
            break
        elif diff / mean < tol and diff < tol:
            break
    return mean
